from s2rcp.core.exceptions import S2rcpDecodeError
from s2rcp.core.container import Container
from s2rcp.core.commands import get_command_by_id


def decode(data):
    """
    Decodes an array of bytes by S2RCP.
    Return:
     - Container object
     - Number of bytes processed
    """
    time = (data[0] << 4) + (data[1] >> 4)
    commands_num = data[1] & 0xF 
    processed = commands_num * 2 + 2
    commands = list()
    for i in range(2, processed, 2):
        command_type = data[i] & 0x2
        command_class = get_command_by_id(command_type)
        if not command_class:
            raise S2rcpDecodeError(
                "unknown command type: {t}"
                .format(t=bin(command_type))
            )
        command_decoder = command_class.get_decoder()
        command = command_decoder.decode(data[i:i+2])
        commands.append(command)
    container = Container(time, commands)
    return container, processed



import s2rcp.core.utils
from s2rcp.network import ClientEvent
from s2rcp.core.commands import (
    StartCommand,
    StopCommand
)

from .motor import BaseMotor
from . import decoder as s2rcp_decoder


class MotorsController:
    def __init__(self, network_client):
        self._motors = dict()
        
        self._client = network_client
        self._client.subscribe(ClientEvent.MESSAGE, self._on_message)

        self._runned = False
        self._ignore_latency = True
        self._max_latency = None

    def set_motor(self, id, motor):
        assert issubclass(type(motor), BaseMotor)
        self._motors[id] = motor

    def start(self):
        self._runned = True

    def stop(self):
        self._runned = False
        for motor in self._motors.values():
            motor.stop()

    def _on_message(self, data):
        """
        Receiving a message, decoding and executing commands
        """
        if not self._runned:
            return

        containers = self._decode_recursively(data)
        for container in containers:

            # Get time last 12 bits to calculate latency
            now = s2rcp.core.utils.get_time_ms_12bits()
            
            latency = now - container.time
            if not self._ignore_latency and latency > self._max_latency:
                continue

            for command in container.commands:
                self._execute_command(command)

    def _decode_recursively(self, data):
        """
        It decodes using the S2RCP protocol, and if there is still information
        left after decoding one container, it tries to decode the remaining
        information. I.e. if the server sent two containers in one message
        (this is possible in TCP), the method decodes them both
        """
        containers = list()
        container, processed = s2rcp_decoder.decode(data)
        containers.append(container)

        if len(data) > processed:
            # Continue decoding
            containers.extend(self._decode_recursively(data[processed:]))
        return containers

    def _execute_command(self, command):
        motor = self._motors.get(command.motor_id, None)
        if not motor:
            # This command is addressed to another controller
            return

        if type(command) is StartCommand:
            motor.start(
                command.speed,
                command.inverted
            )
        elif type(command) is StopCommand:
            motor.stop()



from threading import Thread
from abc import ABCMeta, abstractmethod
from enum import Enum


class ClientEvent(Enum):
    START = 1
    MESSAGE = 2
    STOP = 3


class BaseClient(metaclass=ABCMeta):
    def __init__(self):
        self._handlers = dict()

    @abstractmethod
    def send(self, data) -> None:
        pass   

    def subscribe(self, event_type, handler) -> None:
        """
        Subscribes a function to a specific event.
        When calling a function, arguments are passed as a dictionary
        """
        if event_type not in self._handlers.keys():
            self._handlers[event_type] = list()
        self._handlers[event_type].append(handler)

    def _call_handlers(self, event_type, *args):
        """
        Calls all handlers subscribed to this event
        When calling a handler, arguments are passed as a dictionary
        """
        if event_type not in self._handlers.keys():
            return

        for handler in self._handlers[event_type]:
            Thread(target=handler, args=args, daemon=True).start() 




from threading import Thread
from socket import socket, AF_INET, SOCK_STREAM, SHUT_RDWR

from .client import BaseClient, ClientEvent


class TcpClient(BaseClient):
    def __init__(self):
        super().__init__()
        self._runned = False
        self._socket = socket(AF_INET, SOCK_STREAM)

    def connect(self, server_address):
        if self._runned:
            raise Exception("TcpClient already runned")
        self._socket.connect(server_address)
        self._start_loop()

    def listen(self, address):
        if self._runned:
            raise Exception("TcpClient already runned")
        self._socket.bind(address)
        self._socket, _ = self._socket.accept()
        self._start_loop()

    def send(self, data):
        if not self._runned:
            return
        self._socket.send(data)

    def stop(self):
        if not self._runned:
            return
        self._runned = False
        self._call_handlers(ClientEvent.STOP)
        try:
            self._socket.shutdown(SHUT_RDWR)
        except OSError:
            pass

    def _start_loop(self):
        self._runned = True
        Thread(
                name = "LISTEN_LOOP",
                target = self._loop,
                daemon = True
        ).start()
        self._call_handlers(ClientEvent.START)

    def _loop(self):
        """
        Listens for a connection to the server. When receiving a message 
        from it, calls handlers subscribed to ON_MESSAGE
        """
        while self._runned:
            data = self._socket.recv(128)
            # If the data is empty, it's a disconnect signal
            if not data:
                break
            self._call_handlers(ClientEvent.MESSAGE, data)

        self.stop()





from .command import (
    Command, CommandEncoder, CommandDecoder
)
from .all_commands import (
    get_command_by_id, get_command_type_id
)
from .start_command import StartCommand
from .stop_command import StopCommand


__all__ = [
    "Command",
    "CommandEncoder",
    "CommandDecoder",
    "StartCommand",
    "StopCommand",
    "get_command_by_id",
    "get_command_type_id",
]




import s2rcp.core.utils
from s2rcp.core.exceptions import S2rcpEncodeError


def encode(commands):
    """
    Wraps commands in a container and encodes according to S2RCP
    """
    if not commands:
        raise S2rcpEncodeError("commands list must be not empty")
    time = s2rcp.core.utils.get_time_ms_12bits()
    commands_num = len(commands)
    encoded = bytearray([
        time >> 4,
        ((time & 0xF) << 4) + commands_num
    ])
    for command in commands:
        encoder = command.get_encoder()
        encoded += encoder.encode(command)
    return encoded






class RobotModel():
    def __init__(self, axes_config):
        self._axes_config = axes_config
        self._motors = dict.fromkeys(
                self._axes_config.get_all_motors(), 0.0
        )

    def set_axis_value(self, axis_name, value):
        if abs(value) > 1.0:
            raise ValueError("axis value must be between -1.0 and +1.0")
        
        motors_dict = self._axes_config.get_motors_from_axis(axis_name)
        for motor_id, k in motors_dict.items():
            motor_value = self._motors[motor_id] + k * value
            if motor_value < -1.0:
                self._motors[motor_id] = -1.0
            elif motor_value > 1.0:
                self._motors[motor_id] = 1.0
            else:
                self._motors[motor_id] = motor_value

    def get_all_motors_id(self):
        return list(self._motors.keys())

    def get_motor_value(self, motor):
        return self._motors[motor]



from abc import ABCMeta, abstractmethod


class BaseMotor(metaclass=ABCMeta):
    @abstractmethod
    def start(self, speed, inverted):
        pass

    @abstractmethod
    def stop(self):
        pass



from s2rcp.core.exceptions import S2rcpEncodeError

from .command import Command, CommandEncoder, CommandDecoder
from .all_commands import (
    get_command_type_id, register_command_type
)


class StopCommand(Command):
    def __init__(self, motor_id):
        super().__init__(motor_id)
        self.command_type = get_command_type_id(self.__class__)

    @staticmethod
    def get_encoder():
        return StopCommandEncoder()

    @staticmethod
    def get_decoder():
        return StopCommandDecoder() 


class StopCommandEncoder(CommandEncoder):
    def encode(self, command):
        if type(command) is not StopCommand:
            raise S2rcpEncodeError(
                "type of command must be StopCommand; {t} given"
                .format(t=type(command))
            )
        return bytearray([
            (command.motor_id << 2) + command.command_type,
            0       # Empty byte
        ])


class StopCommandDecoder(CommandDecoder):
    def decode(self, data):
        return StopCommand(
            motor_id = data[0] >> 2
        )


register_command_type(StopCommand, 0b10)




from s2rcp.core.commands import (
    StartCommand, StopCommand
)

from .robot_model import RobotModel
from . import encoder as s2rcp_encoder


class RemoteController:

    def __init__(self, network_client, axes_config):
        self._axes_config = axes_config
        self._client = network_client
        self._model = RobotModel(axes_config)
        self._buffered_model = RobotModel(axes_config)

    def set_axis_value(self, axis_name, value):
        self._buffered_model.set_axis_value(axis_name, value)

    def update(self):
        commands = list()
        for motor_id in self._model.get_all_motors_id():
            actual_value   = self._model.get_motor_value(motor_id)
            buffered_value = self._buffered_model.get_motor_value(motor_id)
            if actual_value != buffered_value:
                if buffered_value == 0.0:
                    command = self._create_stop_command(motor_id)
                else:
                    command = self._create_start_command(motor_id, buffered_value)
                commands.append(command)

        self._reset_buffer()
        self._send_commands(commands)

    def _reset_buffer(self):
        self._model = self._buffered_model
        self._buffered_model = RobotModel(self._axes_config)

    def _send_commands(self, commands):
        encoded_commands = s2rcp_encoder.encode(commands)
        self._client.send(encoded_commands)

    def _create_start_command(self, motor_id, speed):
        if abs(speed) > 1.0:
            raise ValueError(
                "speed must be in range -1.0 .. +1.0 ({given} given)"
                .format(given=speed)
            )

        speed = abs(int(speed * StartCommand.MAX_SPEED))
        inverted = int(speed < 0.0)
        
        return StartCommand(motor_id, speed, inverted)

    def _create_stop_command(self, motor_id):
        return StopCommand(motor_id)



from time import time


def check_int_value_is_valid(value, min, max):
    if not _is_int_value_valid(value, min, max):
        raise ValueError("value must be in range [{min}; {max}] ({value} given)"
                           .format(min=min, max=max, value=value))


def _is_int_value_valid(value, min, max):
    if type(value) is not int:
        return False
    return value >= min and value <= max


def get_time_ms_12bits():
    return int(time() * 1000) & 0xFFF



from s2rcp.core.utils import check_int_value_is_valid
from s2rcp.core.exceptions import S2rcpEncodeError

from .command import Command, CommandEncoder, CommandDecoder
from .all_commands import (
    get_command_type_id, register_command_type
)


class StartCommand(Command):

    MAX_SPEED = 2**7-1

    def __init__(self, motor_id, speed, inverted):
        super().__init__(motor_id)
        self.command_type = get_command_type_id(self.__class__)
        check_int_value_is_valid(speed, min=0, max=StartCommand.MAX_SPEED)
        self.speed = speed
        self.inverted = inverted

    def get_max_speed(self):
        return self.MAX_SPEED

    @staticmethod
    def get_encoder():
        return StartCommandEncoder()

    @staticmethod
    def get_decoder():
        return StartCommandDecoder()


class StartCommandEncoder(CommandEncoder):
    def encode(self, command):
        if type(command) is not StartCommand:
            raise S2rcpEncodeError(
                "type of command must be StartCommand; {t} given"
                .format(t=type(command))
            )
        return bytearray([
            (command.motor_id << 2) + command.command_type,
            (command.inverted << 7) + command.speed
        ])


class StartCommandDecoder(CommandDecoder):
    def decode(self, data):
        return StartCommand(
            motor_id = data[0] >> 2,
            speed    = data[1] & 0x7F,
            inverted = bool(data[1] >> 7)
        )


register_command_type(StartCommand, 0b01)




from .utils import check_int_value_is_valid


class Container:
    def __init__(self, time, commands):
        check_int_value_is_valid(time, min=0, max=2**12-1)
        self.time = time
        self.commands = commands




from abc import ABCMeta, abstractmethod
from s2rcp.core.utils import check_int_value_is_valid


class Command(metaclass=ABCMeta):
    def __init__(self, motor_id):
        check_int_value_is_valid(motor_id, min=0, max=2**6-1)
        self.motor_id = motor_id

    @staticmethod
    @abstractmethod
    def get_encoder():
        pass

    @staticmethod
    @abstractmethod
    def get_decoder():
        pass


class CommandEncoder(metaclass=ABCMeta):
    @abstractmethod
    def encode(self, command):
        pass


class CommandDecoder(metaclass=ABCMeta):
    @abstractmethod
    def decode(self, data):
        pass




import setuptools

with open("README.md") as file:
    read_me_description = file.read()

setuptools.setup(
    name="s2rcp",
    version="1.0.0",
    author="zhikh",
    author_email="zhikh.k@gmail.com",
    description="Simple Remote Robot Control Protocol",
    long_description=read_me_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/zhikh/s2rcp",
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)



from s2rcp.core.exceptions import S2rcpException


_map_command_to_id = dict()
_map_id_to_command = dict()


def register_command_type(command_type, command_type_id):
    if command_type_id in _map_id_to_command:
        raise S2rcpException(
            "command with id={id_} has already been registered"
            .format(id_=command_type_id)
        )
    _map_command_to_id[command_type] = command_type_id
    _map_id_to_command[command_type_id] = command_type


def get_command_type_id(command_type):
    return _map_command_to_id.get(command_type, None)


def get_command_by_id(command_type_id):
    return _map_id_to_command.get(command_type_id, None)




import json


class AxesConfig:

    def __init__(self):
        self._axes = dict()

    def setup_from_json(self, json_str):
        self._axes = json.loads(json_str)

    def add_new_axis(self, axis_name):
        self._axes[axis_name] = dict()

    def add_motor_to_axis(self, axis_name, motor_id, motor_k):
        if axis_name not in self._axes.keys():
            self.add_new_axis(axis_name)
        self._axes[axis_name][motor_id] = motor_k

    def get_all_axes(self):
        return list(self._axes.keys())

    def get_motors_from_axis(self, axis_name):
        return self._axes[axis_name]
    
    def get_all_motors(self):
        all_motors = list()
        for axis_name in self._axes:
            all_motors.extend(self._axes[axis_name].keys())
        return all_motors
